# AUTOGENERATED FILE! PLEASE DON'T EDIT
import k1lib
import k1lib.cli as cli
try:
    from rdkit import Chem
    from rdkit.Chem import Draw
    from rdkit.Chem import AllChem
    from rdkit.Chem.Draw import IPythonConsole
    mol_available = True
    IPythonConsole.drawOptions.addAtomIndices = True
    #__all__ = ["toMol", "toSmiles"]
    __all__ = []
except: mol_available = False; __all__ = []
if mol_available:
    def toMol():
        """Smiles to molecule.
Example::

    "c1ccc(C)cc1" | toMol()"""
        return cli.aS(Chem.MolFromSmiles)
    def toSmiles():
        """Molecule to smiles.
Example::

    "c1ccc(C)cc1" | toMol() | toSmiles()"""
        return cli.aS(Chem.MolToSmiles)